## Meta Sepia
#
# location.rb
# - provides class to manage current and previous locations
# - provides methods to manipulate the location
# - provides methods to retrieve its history in some formats
#
# author::    Peter Rullmann (mailto:peter@p4n.net)
# copyright:: Copyright (c) 2006-2008 Peter Rullmann
#
# You can Freely distribute this program under the GNU General Public
# License. See the file "COPYING" for the exact licensing terms.


require 'gtk2'
require 'date'




class Location < GLib::Object
  attr_reader :loc_obj
  type_register
  signal_new("changed", GLib::Signal::RUN_FIRST, nil, nil)
  def signal_do_changed(); end

  FORGET_FACTOR = 1.2
  FORGET_TERM = 1
  MAX_IMPRESSIONS = 10
  REMEMBER_BACK = 20
  CONFIG_PATH = ENV["HOME"] + "/.filebrowser"
  MOUNT_EXCLUDES = ["/proc", "/sys", "/var", "/dev", "/lib"]
  MOUNT_EXCLUDES.collect! { |str| Regexp.new("^" + str) }


  def initialize(path)
    super(nil)
    @path = path
    @loc_obj = LocationObject.new(path)
    @old_locs = {path => @loc_obj}
    @history = [ path ]
    @histpos = 0
    @last_places = [ path ]
    @favorites = LocationHistory.load
    
    $glade["up_action"].signal_connect("clicked") do |se|
      self.path= File.dirname(@loc_obj.path)
    end
    $glade["back_action"].signal_connect("clicked") do |se|
      if @histpos > 0
        @histpos -= 1
        path = @history[@histpos]
        @loc_obj = @old_locs[ path ]
        @last_places.unshift path
        @favorites.memorise(@last_places, path)
        signal_emit "changed"
      end
    end
    $glade["forward_action"].signal_connect("clicked") do |se|
      if @histpos < @history.size-1
        @histpos += 1
        path = @history[@histpos]
        @loc_obj = @old_locs[ path ]
        @last_places.unshift path
        @favorites.memorise(@last_places, path)
        signal_emit "changed"
      end
    end

  end


  def path=(path)
    path = path.gsub(%r{//+}, "/").sub(%r{/$}, "")
    path = "/" if path == ""
    return if path == @loc_obj.path
    @favorites.memorise(@last_places, path)
    @last_places.unshift path
    if @old_locs.has_key? path
      @loc_obj = @old_locs[path]
    else
      @loc_obj = LocationObject.new(path)
    end
    @histpos += 1
    @history = @history[0, @histpos]
    @history << path
    signal_emit "changed"
  end


  def system_places
    return @system_places unless @system_places.nil?
    # possible enhancement: wait for file change on /etc/mtab
    mounts = IO.readlines("/etc/mtab").collect { |line|
      line =~ /^[^ ]+ ([^ ]+) /
      mount = $1
      MOUNT_EXCLUDES.each { |regex| mount = nil if mount =~ regex }
      mount
    }
    mounts.compact!

    @system_places = Hash.new("folder")
    mounts.each { |mnt| @system_places[mnt]= "drive-harddisk" }
    home = ENV["HOME"]
    presets = {"/" => "computer", home => "gnome-fs-home", home +
               "/Desktop" => "desktop", home + "/.Trash" => "emptytrash"}
    @system_places.update presets
  end


  def important_places(rows_num)
    # goal: find a list of paths p out of four lists of paths in r where:
    #       * for each path x = "/a/b/c" out of p there has to exist a
    #         path y = "/a/b" and
    #       * the size of p should be rows_num but not more and
    #       * the first items of r[x] where x = 1..4 are to be prioritized
    # solution: - loop over the elements r and try to fit an element in p,
    #             starting from the top of r[x] where x = 1..4.
    #           - if it fits, put it in, if not, try the next x
    icons = system_places.dup.delete_if { |path, icon| icon == "folder" }
 
    # first priority: currend directory and its parent directories
    rows = [parent = @loc_obj.path]
    rows << parent = File.dirname(parent) until parent == "/"

    if rows.size >= rows_num
      return Hash[ *((rows.collect { |r| [r, icons[r]] }).flatten) ]
    end

    # second priority: system places
    system_places.keys.each do |path|
      # find all parents of path
      parents = [parent = path]
      parents << parent = File.dirname(parent) until parent == "/"
      # add path and its parents to rows if rows isn't be to big afterwords
      union = rows | parents
      rows = union unless union.size > rows_num
      break if rows.size == rows_num
    end

    if rows.size >= rows_num
      return Hash[ *((rows.collect { |r| [r, icons[r]] }).flatten) ]
    end

    # third priority: last places and places most visited from currend path
    groups = [@last_places,
              @favorites.most_visited_from(@loc_obj.path)]
    sizes = groups.collect { |arr| arr.size }
    # iterate over paths in this order:
    # group[1][1], group[2][1], group[1][2], group[2][2], group[1][3], ...
    0.upto(sizes.max-1) do |place|
      groups.each_index do |group|
        next unless place < sizes[group]
        path = groups[group][place]
        next if path == ""
        # find all parents of path
        parents = [parent = path]
        parents << parent = File.dirname(parent) until parent == "/"
        # add path and its parents to rows if rows isn't be to big afterwords
        union = rows | parents
        rows = union unless union.size > rows_num
        break if rows.size == rows_num
      end
      break if rows.size == rows_num
    end

    if rows.size >= rows_num
      return Hash[ *((rows.collect { |r| [r, icons[r]] }).flatten) ]
    end

    # fourth priority: most visited places
    @favorites.most_visited.each do |path|
      # find all parents of path
      parents = [parent = path]
      parents << parent = File.dirname(parent) until parent == "/"
      # add path and its parents to rows if rows isn't be to big afterwords
      union = rows | parents
      rows = union unless union.size > rows_num
      break if rows.size == rows_num
    end

    return Hash[ *((rows.collect { |r| [r, icons[r]] }).flatten) ]
  end


  private

  class LocationObject
    attr_accessor :path, :model, :view_type, :icon_size, :num_lines,
                  :show_hidden, :scroll_point, :sort_by, :reverse_sort,
                  :folders_first, :group_by, :visible_columns, :model_columns
    def initialize(path)
      @path = path
      @model = nil; @view_type = :list; @icon_size = :small; @num_lines = 1
      @show_hidden = false; @scroll_point = 0; @sort_by = "File:Name"
      @reverse_sort = false; @folders_first = true; @group_by = nil
      @visible_columns = ["File:Name", "File:Type", "File:Size"]
      @model_columns = []
    end
  end


  class LocationHistory

    def initialize
      # example: { :date => Date.today, "/home" => 21, "/bin" => 3 }
      @lastplaces = {}
      # example: { "/home" => { :date => Date.today, "/home/peterpan" => 14 } }
      @followingplaces = {}
    end

    def LocationHistory.load()
      return LocationHistory.new if !FileTest.exists?(CONFIG_PATH + "/history")
      File.open(CONFIG_PATH + "/history") do |file|
        return ret = Marshal.load(file)
      end
    end

    def save
      Dir.mkdir(CONFIG_PATH) unless FileTest.exists? CONFIG_PATH
      File.open(CONFIG_PATH + "/history", "w") do |file|
        Marshal.dump(self, file)
      end
    end

    def most_visited
      (lastplaces = @lastplaces.clone).delete(:date)
      return (lastplaces.sort { |a,b| b[1]<=>a[1] }).collect { |a| a[0] }
    end
    
    def most_visited_from(path)
      return [] unless @followingplaces.has_key? path
      (places = @followingplaces[path].clone).delete(:date)
      return (places.sort { |a,b| b[1]<=>a[1] }).collect { |a| a[0] }
    end

    # remember the given new_path in the global @lastplaces history and in the
    # history of each previous visited path in @followingplaces[last_paths[i]]
    def memorise(last_paths, new_path)
      increase_path(new_path, @lastplaces)
      last_paths[0..REMEMBER_BACK].each do |last_path|
        @followingplaces[last_path] = {} if @followingplaces[last_path].nil?
        increase_path(new_path, @followingplaces[last_path])
      end
      save
    end

    private

    def increase_path(path, places)
      LocationHistory.forget places
      old_count = (places.has_key?(path) ? places[path] : 0)
      places[path] = [old_count + 1, MAX_IMPRESSIONS].min
    end      

    def LocationHistory.forget(places)
      return if places[:date] == Date.today
      places.each_key do |key|
        next if key == :date
        new_value = (places[key] / FORGET_FACTOR).to_i - FORGET_TERM
        new_value > 0 ? places[key] = new_value : places.delete(key)
      end
      places[:date] = Date.today
    end

  end


end




$location = Location.new(ENV["HOME"])

