## Meta Sepia
#
# metadatatypes.rb
# provides basic metadata types and widgets to show and edit these
#
# author::    Peter Rullmann (mailto:peter@p4n.net)
# copyright:: Copyright (c) 2006-2008 Peter Rullmann
#
# You can Freely distribute this program under the GNU General Public
# License. See the file "COPYING" for the exact licensing terms.


require 'gtk2'
require 'date'





###   ##   ### ####   ###  ###   ###  ###   ###   #   # ### ###   ##  #### ##### 
#  # #  # #    #      #  # #  # #   # #  # #      #   #  #  #  # #    #      #   
###  ####  ##  ###    ###  ###  #   # ###   ##    #   #  #  #  # # ## ###    #   
#  # #  #    # #      #    #  # #   # #       #   # # #  #  #  # #  # #      #   
###  #  # ###  ####   #    #  #  ###  #    ###     # #  ### ###   ### ####   #   
class BasePropsWidget < Gtk::EventBox
  type_register
  signal_new("changed", GLib::Signal::RUN_FIRST, nil, nil)
  def signal_do_changed()
    values_changed(*@values)
    value_changed(*@values)
  end

  def initialize(title, editable, *values)
    super()
    @editable = editable
    @values = values
    @pointer_in = false

    add @hbox = Gtk::HBox.new(false, 0)
    # show title only when a title was given
    label = Gtk::Label.new(" #{title}:")
    @hbox.pack_start(label, false) unless title.to_s == ""

    @disp_widgets = []
    @edit_widgets = []
    @size_groups = {}
    init_disp(*@values)
    init_edit(*@values) if @editable
    init_hover_effect() if @editable
  end


  def init_disp(*values)
    # this method is to be overridden by inheriting class
  end

  def init_edit(*values)
    # this method may be overridden by inheriting class
  end
  
  def value_changed(*values); end
  def values_changed(*values)
    # one of these two methods is to be overridden by inheriting class
  end


  def set_value(*values)
    set_values(*values)
  end
  def set_values(*values)
    if @values != values
      @values = values
      signal_emit "changed"
    end
  end

  def value
    return @values[0]
  end
  def values
    return @values
  end


  def init_hover_effect()
    # show editor widgets only when cursor is over widget or one has focus
    hide_edit
    @edit_widgets.each do |edit_widget|
      edit_widget.signal_connect("focus-out-event") do |arg1, arg2|
        hide_edit if !@pointer_in
        false
      end
      edit_widget.signal_connect("focus-in-event") do |arg1, arg2|
        show_edit
        false
      end
    end
    signal_connect("enter-notify-event") do |arg1, arg2|
      if arg2.detail != Gdk::EventCrossing::NOTIFY_INFERIOR and
         arg2.mode == Gdk::EventCrossing::NORMAL
        @pointer_in = true
	show_edit
      end
    end
    signal_connect("leave-notify-event") do |arg1, arg2|
      if arg2.detail != Gdk::EventCrossing::NOTIFY_INFERIOR and 
         arg2.mode == Gdk::EventCrossing::NORMAL
        @pointer_in = false
        if !@edit_widgets.detect { |edit_widget| edit_widget.has_focus? }
	  hide_edit
        end
      end
    end
  end
  
  def show_edit
    @disp_widgets.each { |disp_widget| disp_widget.visible= false }
    @edit_widgets.each { |edit_widget| edit_widget.visible= true }
  end
  def hide_edit
    @disp_widgets.each { |disp_widget| disp_widget.visible= true }
    @edit_widgets.each do |edit_widget|
      edit_widget.visible= false
      edit_widget.no_show_all= true
    end
  end
  
  
  # wrapper methods for adding widgets to hbox
  # TODO: new scheme: add_mutual(disp_wid, edit_wid), add_greying(edit_wid)
  def pack(child, size_group = nil)
    if !size_group.nil?
      if !@size_groups.has_key?(size_group)
        @size_groups[size_group] = Gtk::SizeGroup.new(Gtk::SizeGroup::BOTH)
        @size_groups[size_group].ignore_hidden= false
      end
      @size_groups[size_group].add_widget child
    end
    @hbox.pack_start(child)
  end

  def pack_disp(child, size_group = nil)
    pack(child, size_group)
    @disp_widgets << child
  end

  def pack_edit(child, size_group = nil)
    pack(child, size_group)
    @edit_widgets << child
  end
  def add_edit(child)
    @edit_widgets << child
  end
end





###   ##  ##### ####   #   # #### #####  ##    ##### #   # ###  #### 
#  # #  #   #   #      ## ## #      #   #  #     #   #   # #  # #    
#  # ####   #   ###    # # # ###    #   ####     #    # #  ###  ###  
#  # #  #   #   #      #   # #      #   #  #     #     #   #    #    
###  #  #   #   ####   #   # ####   #   #  #     #     #   #    #### 
class DateMetaType
  def DateMetaType.tooltip(title, time)
    return time.strftime "#{title}: %a %b %d, %H:%M"
  end
  
  
  def DateMetaType.display(time)
    today = Date.today
    diff = today - Date.new(time.year, time.mon, time.mday)
    if diff < 1
      return time.strftime "Today, %H:%M"
    elsif diff < 2
      return time.strftime "Yesterday, %H:%M"
    elsif diff < 7
      return time.strftime "%A, %H:%M"
    else
      return time.strftime "%a %m/%d/%Y, %H:%M"
    end
  end
  
  
  def DateMetaType.ruby_class; return Time; end
  def DateMetaType.tracker_to_ruby(tracker_str)
    # TODO: what format is date in tracker?
    tracker_str
  end
  def DateMetaType.ruby_to_tracker(ruby_time)
    ruby_time.to_s
  end
  
  
  class PropsWidget < BasePropsWidget
    def value_changed(time)
      @ddisp.text = time.strftime "%m/%d/%Y"
      @tdisp.text = time.strftime "%H:%M:%S"
    end

    def init_disp(time)
      # hack, so that he can compare correctly if time has changed
      @values[0] = Time.local(*time.to_a)

      # display date
      @ddisp = Gtk::Label.new(time.strftime "%m/%d/%Y")
      @ddisp.set_alignment(0, 0.5)
      @ddisp.set_padding(5, 0)
      pack_disp(@ddisp)
      
      # display time
      @tdisp = Gtk::Label.new(time.strftime "%H:%M:%S")
      @tdisp.set_alignment(0, 0.5)
      @tdisp.set_padding(5, 0)
      pack_disp(@tdisp)
    end

    def init_edit(time)
      # setup the date editor
      dedit = Gnome::DateEdit.new(time, false, false)
      # date editor and date label should have the same size request
      w = [@ddisp.size_request[0], dedit.size_request[0]].max + 4
      h = [@ddisp.size_request[1], dedit.size_request[1]].max
      @ddisp.set_size_request(w, h)
      dedit.set_size_request(w, h)
      dedit.visible= false
      dedit.no_show_all= true
      pack_edit(dedit)

      # setup the two SpinButtons to edit the time
      hedit = Gtk::SpinButton.new(0,23,1)
      medit = Gtk::SpinButton.new(0,59,1)
      hedit.value= time.hour
      medit.value= time.min
      # the two editors and the label should have the same size request
      w = [@tdisp.size_request[0], 42 * 2].max + 4
      h = [@tdisp.size_request[1], hedit.size_request[1]].max
      @tdisp.set_size_request(w, h)
      hedit.set_size_request(w / 2, h)
      medit.set_size_request(w - w / 2, h)
      hedit.visible=     medit.visible=     false
      hedit.no_show_all= medit.no_show_all= true
      pack_edit(hedit)
      pack_edit(medit)

      # change value if time or date is edited
      dedit.signal_connect("date-changed") do |arg1, arg2|
        # replace self.value with new time with just the date-part changed
        time_arr = self.value.to_a[0, 3] + Time.at(dedit.time).to_a[3, 7]
        set_value Time.local(*time_arr)
      end
      hedit.signal_connect("value-changed") do
        # replace self.value with new time with just the hour-part changed
        (time_arr = self.value.to_a)[2] = hedit.value
        set_value Time.local(*time_arr)
      end
      medit.signal_connect("value-changed") do
        # replace self.value with new time with just the minute-part changed
        (time_arr = self.value.to_a)[1] = medit.value
        set_value Time.local(*time_arr)
      end
    end
  end
end





 ### ##### ###  ### #   #  ##    #   # #### #####  ##    ##### #   # ###  #### 
#      #   #  #  #  ##  # #      ## ## #      #   #  #     #   #   # #  # #    
 ##    #   ###   #  # # # # ##   # # # ###    #   ####     #    # #  ###  ###  
   #   #   #  #  #  #  ## #  #   #   # #      #   #  #     #     #   #    #    
###    #   #  # ### #   #  ###   #   # ####   #   #  #     #     #   #    #### 
class StringMetaType
  def StringMetaType.tooltip(title, text)
    return "#{title}: #{text}"
  end
  
  
  def StringMetaType.display(text)
    return text
  end
  
  def StringMetaType.ruby_class; return String; end

  class PropsWidget < BasePropsWidget
    def init_disp(text)
      # display text
      @disp = Gtk::Label.new(text)
      @disp.set_alignment(0, 0.5)
      @disp.set_padding(5, 0)
      pack_disp @disp, :text
    end      
      
    def init_edit(text)
      # setup the text editor
      edit = Gtk::Entry.new
      edit.text= text
      pack_edit edit, :text

      edit.signal_connect("activate") do |arg1|
        set_value edit.text
        # this is a hack to loose focus
        # please tell me how this is done properly --peterpan
        edit.visible= false
        edit.visible= true if @pointer_in
      end
    end
    
    def value_changed(text)
      @disp.text= text
    end
  end
end





#### ### #    ####  ### ### ##### ####   #   # #### #####  ##          
#     #  #    #    #     #     #  #      ## ## #      #   #  #         
###   #  #    ###   ##   #    #   ###    # # # ###    #   ####         
#     #  #    #       #  #   #    #      #   # #      #   #  #         
#    ### #### #### ###  ### ##### ####   #   # ####   #   #  #   # # # 
class FilesizeMetaType
  def FilesizeMetaType.tooltip(title, size)
    return "#{title}: #{size} bytes"
  end
  
  
  def FilesizeMetaType.display(size)
    str = size.to_s
    units = ["Byte", "KiB", "MiB", "GiB", "TiB", "PiB"]
    quotient, modulus = (str.size - 1).divmod 3
    dot = quotient > 0 ? "." : ""
    str[0, modulus+1] + dot + str[modulus+1, 3-modulus] + " " + units[quotient]
  end
  
  def StringMetaType.ruby_class; return Integer; end
  
  class PropsWidget < BasePropsWidget
    def init_disp(size)
      # display text
      @disp = Gtk::Label.new(FilesizeMetaType.display(size))
      @disp.set_alignment(0, 0.5)
      @disp.set_padding(5, 0)
      pack_disp @disp

      # no editing possible
      @editable = false
    end      
      
    def value_changed(size)
      @disp.text= FilesizeMetaType.display(size)
    end
  end
end





### #   #  ##   ##  ####  ### ### ##### ####   #   # #### #####  ## 
 #  ## ## #  # #    #    #     #     #  #      ## ## #      #   #  #
 #  # # # #### # ## ###   ##   #    #   ###    # # # ###    #   ####
 #  #   # #  # #  # #       #  #   #    #      #   # #      #   #  #
### #   # #  #  ### #### ###  ### ##### ####   #   # ####   #   #  #   # # #
class ImagesizeMetaType
  def ImagesizeMetaType.tooltip(title, width, height)
    return "#{title}: width: #{width.to_s}px, height: #{height}px"
  end
  
  
  def ImagesizeMetaType.display(width, height)
    return "#{width.to_s}x#{height.to_s}px"
  end
  
  def ImagesizeMetaType.ruby_class; return [Integer, Integer]; end
  
  class PropsWidget < BasePropsWidget
    def init_disp(width, height)
      # display text
      @disp = Gtk::Label.new(ImagesizeMetaType.display(width, height))
      @disp.set_alignment(0, 0.5)
      @disp.set_padding(5, 0)
      pack_disp @disp

      # no editing possible
      @editable = false
    end      
      
    def value_changed(width, height)
      @disp.text= ImagesizeMetaType.display(width, height)
    end
  end
end





###   ##  #   # #  #   #   # #### #####  ##    ##### #   # ###  #### 
#  # #  # ##  # # #    ## ## #      #   #  #     #   #   # #  # #    
###  #### # # # ##     # # # ###    #   ####     #    # #  ###  ###  
#  # #  # #  ## # #    #   # #      #   #  #     #     #   #    #    
#  # #  # #   # #  #   #   # ####   #   #  #     #     #   #    #### 
class RankMetaType
  def RankMetaType.tooltip(title, stars)
    return "#{title}: #{stars.to_s} stars"
  end
  
  
  def RankMetaType.display(stars)
    return "*" * stars
  end
  
  def RankMetaType.ruby_class; return Integer; end
  
  # arrays of prerendered stars for ranking purposes
  RED_STARS, YELL_STARS = ["stars_red.gif", "stars_yell.gif"].collect do |color|
    (0..10).to_a.collect do |n|
      back = Gdk::Pixbuf.new("icons/stars_back.gif")
      full = Gdk::Pixbuf.new("icons/" + color)
      back = back.scale(76, 13); full = full.scale(76, 13)
      full.copy_area(0, 0, full.width*n/10, full.height, back, 0, 0)
      back
    end
  end
  
  
  class PropsWidget < BasePropsWidget
    def init_disp(stars)
      @img = Gtk::Image.new(YELL_STARS[stars])
      @img.set_alignment(0, 0.5)
      @img.set_padding(5, 0)
      pack_disp @img

      add_events(Gdk::Event::POINTER_MOTION_MASK)
      signal_connect("motion-notify-event") do |se, ev|
        width = YELL_STARS[0].width; x = @img.pointer[0]
        if x < 0 or x > width + 10
          (@img.pixbuf= YELL_STARS[value]; @pointer_in = false) if @pointer_in
        else
          @pointer_in = true
          @img.pixbuf= RED_STARS[ [10*x/width,10].min ]
        end
      end
      signal_connect("leave-notify-event") do
	@img.pixbuf= YELL_STARS[value]
      end
      signal_connect("button-release-event") do |se, ev|
        width = YELL_STARS[0].width; x = @img.pointer[0]
        if x > 0 and x < width + 10
          set_value [10*x/width,10].min
        end
      end
      
      # I have handled editing myself
      @editable = false
    end
    
    def value_changed(stars)
      @img.pixbuf= YELL_STARS[stars]
    end
  end
end





##### #### #   # #####   #   # #### #####  ##    ##### #   # ###  #### 
  #   #     # #    #     ## ## #      #   #  #     #   #   # #  # #    
  #   ###    #     #     # # # ###    #   ####     #    # #  ###  ###  
  #   #     # #    #     #   # #      #   #  #     #     #   #    #    
  #   #### #   #   #     #   # ####   #   #  #     #     #   #    #### 
class TextMetaType
  def TextMetaType.tooltip(title, text)
    return "#{title}: #{text}"
  end
  
  
  def TextMetaType.display(text)
    return text
  end
  
  def TextMetaType.ruby_class; return String; end

  class PropsWidget < BasePropsWidget
    def initialize(title, editable, text)
      # class handles title itself
      @title = title.to_s
      super("", editable, text)
    end

    def init_disp(text)
      # setup the text editor
      scroll = Gtk::ScrolledWindow.new
      scroll.set_policy Gtk::POLICY_NEVER, Gtk::POLICY_AUTOMATIC
      scroll.shadow_type= Gtk::SHADOW_IN
      @edit = Gtk::TextView.new
      @edit.buffer.text= text
      @edit.left_margin= 4
      @edit.right_margin= 4
      @edit.wrap_mode= Gtk::TextTag::WRAP_WORD
      scroll.add @edit
      pack_edit scroll
      add_edit @edit
      
      @edit.buffer.signal_connect("end-user-action") do |arg1|
        set_value @edit.buffer.text
      end

      @base_color_insensitive = style.base(Gtk::STATE_INSENSITIVE)
      @base_color_normal = style.base(Gtk::STATE_NORMAL)
      @fg_color_insensitive = style.fg(Gtk::STATE_INSENSITIVE)
      @fg_color_normal = style.fg(Gtk::STATE_NORMAL)
    end
    
    # class handles hide and show of editor itself
    def show_edit
      @edit.buffer.text= "" if @no_text
      @edit.modify_base(Gtk::STATE_NORMAL, @base_color_normal)
      @edit.modify_text(Gtk::STATE_NORMAL, @fg_color_normal)
    end
    def hide_edit
      if @no_text = (@edit.buffer.text == "")
        @edit.modify_text(Gtk::STATE_NORMAL, @fg_color_insensitive)
        @edit.buffer.text= "No #{@title}"
      end
      @edit.modify_base(Gtk::STATE_NORMAL, @base_color_insensitive)
    end
    
    def value_changed(text)
      @no_text = false
    end
  end
end







