## Meta Sepia
#
# tracker.rb
# provides api to connect to meta tracker via dbus
# uses direct filesystem calls as backup
#
# author::    Peter Rullmann (mailto:peter@p4n.net)
# copyright:: Copyright (c) 2006-2008 Peter Rullmann
#
# You can Freely distribute this program under the GNU General Public
# License. See the file "COPYING" for the exact licensing terms.

require 'dbus'
require 'gtk2'



class Tracker
  SERVICE   = 'org.freedesktop.Tracker'
  OBJECT    = '/org/freedesktop/tracker'
  INTERFACE = 'org.freedesktop.Tracker'

  def initialize
    # initialize dbus connection to tracker and get all needed interfaces
    @dbus_obj = DBus::SessionBus.instance.service(SERVICE).object(OBJECT)
    @tracker  = DBus::ProxyObjectInterface.new(@dbus_obj, INTERFACE)
    @metadata = DBus::ProxyObjectInterface.new(@dbus_obj, INTERFACE+".Metadata")
    @search   = DBus::ProxyObjectInterface.new(@dbus_obj, INTERFACE+".Search")
    @files    = DBus::ProxyObjectInterface.new(@dbus_obj, INTERFACE+".Files")
    @keywords = DBus::ProxyObjectInterface.new(@dbus_obj, INTERFACE+".Keywords")

    # because tracker does not support dbus introspection yet, we have to define
    # all methods we want to use here
    @tracker.define_method("GetServices", "in b, out a{sv}")
    @tracker.define_method("GetStats", "out aas")
    @metadata.define_method("GetRegisteredClasses", "out as")
    @metadata.define_method("Get", "in s, in s, in as, out as")
    @metadata.define_method("GetTypeDetails", "in s, out s, out b, out b")
    @metadata.define_method("RegisterType", "in s, in s")
    @files.define_method("GetMetadataForFilesInFolder",
                         "in i,in s,in as,out aas")
    @files.define_method("Exists", "in s, in b, out b")
    @keywords.define_method("GetList", "in s, out aas")
    @keywords.define_method("Get", "in s, in s, out aas")
    @keywords.define_method("Search", "in i, in s, in as, in i, in i, out as")
    @search.define_method("Text", "in i, in s, in s, in i, in i, out as")

    # get list of fields and info wether they are writable
    # If a field is a key in @tracker_fields, its value is readable through
    # tracker. If if the value in @tracker_fields, it is also writable
    @metadata.RegisterType("User:Description", "index")
    @tracker_fields = {"User:Description" => true, "User:Keywords" => true}
    skip_fields = @tracker_fields.keys + ["default"]
    @metadata.GetRegisteredClasses[0].each do |f|
      next if skip_fields.include? f
      (data_type, embedded, writable) = @metadata.GetTypeDetails f
      puts "no info about wether #{f} is writable! guess: no" if writable.nil?
      @tracker_fields[f] = writable & true
    end

    # set list of fields of filesystem
    @fs_fields = { "File:Name" => true, "File:Path" => false,
      "File:Size" => false, "File:Permissions" => true, "File:Owner" => false,
      "File:Modified" => false, "File:Accessed" => false,
      "File:Format" => false }
  end


  def get_metadata(file, fields)
    # get metadata from tracker if available there
    tracker_in = @tracker_fields.keys & fields
    tracker_out = {}
    if tracker_in.include? "User:Keywords"
      tracker_in = tracker_in - ["User:Keywords"]
      reply = @keywords.Get("Files", file)
      tracker_out["User:Keywords"] = reply.join(" ")
    end
    if !tracker_in.empty?
      tracker_reply = @metadata.Get("Files", file, tracker_in)[0]
      tracker_in.each_index do |i|
        tracker_out[tracker_in[i]] = tracker_reply[i]
      end
      tracker_in = tracker_out.keys
    end

    # get all metadata that could not be provided by tracker but can be
    # provided by reading the filesystem.
    fs_in = @fs_fields.keys & (fields - tracker_in)
    fs_out = {}
    stat_reply = File.lstat(file) unless
      (fs_in - ["File:Name", "File:Path"]).empty?
    fs_in.each do |field|
      case field
      when "File:Name"
        fs_out[field] = File.basename(file)
      when "File:Path"
        fs_out[field] = file
      when "File:Format"
        h = { "file" => "File", "directory" => "Folder", "link" => "Symlink",
          "characterSpecial" => "Character Device", "socket" => "Socket",
          "blockSpecial" => "Block Device", "fifo" => "Fifo" }
        fs_out[field] = h[stat_reply.ftype]
      when "File:Size"
        fs_out[field] = stat_reply.size
      when "File:Permissions"
        mode = stat_reply.mode
        str = "-rwxrwxrwx"
        8.downto(0) {  |bit| str[9-bit,1] = "-" if mode[bit] == 0 }
        str[9,1]= "T" if mode[9] == 1
        str[6,1]= "S" if mode[10] == 1
        str[3,1]= "S" if mode[11] == 1
        trans = {"file" => "-", "directory" => "d", "characterSpecial" => "c", 
          "blockSpecial" => "b", "fifo" => "p", "link" => "l", "socket" => "s"}
        str[0,1] = trans[stat_reply.ftype]
        fs_out[field] = str
      when "File:Owner"
        fs_out[field] = stat_reply.uid.to_s
      when "File:Modified"
        fs_out[field] = stat_reply.mtime.strftime("%Y-%m-%dT%H:%M:%S+0200")
      when "File:Accessed"
        fs_out[field] = stat_reply.atime.strftime("%Y-%m-%dT%H:%M:%S+0200")
      end
    end

    # merge the data from the two sources and set the writable field
    hash = {}
    tracker_in.each { |f| hash[f]= {:value => tracker_out[f]} }
    fs_in.each      { |f| hash[f]= {:value => fs_out[f]     } }
    hash.each_pair { |f, h| h[:writable]= @tracker_fields[f] || @fs_fields[f] }
    (fields - tracker_in - fs_in).each { |f|
      hash[f]= {:value => nil, :writable => false} }

    hash
  end
  

  def set_metadata(file, hash)
    hash.each_pair do |key, value|
      if key == "User:Keywords"
        new_tags = value.split
        old_tags = @keywords.Get("Files", file)
        to_add = new_tags - old_tags
        to_remove = old_tags - new_tags
        @keywords.Add(   "Files", file, to_add   ) unless to_add.empty?
        @keywords.Remove("Files", file, to_remove) unless to_remove.empty?
      elsif @tracker_fields[key] == true
        @metadata.Set("Files", file, [key], [value])
      elsif key == "File:Name"
        puts "renaming \"#{file}\" to \"#{File.dirname(file) + "/" + value}\""
        File.rename(file, File.dirname(file) + "/" + value)
      elsif key == "File:Permissions"
        mode = 511
      	8.downto(0) { |bit| mode -= 2**bit if value[9-bit,1] == "-" }
        mode += 2** 9 if value[9,1] == "T"
        mode += 2**10 if value[6,1] == "S"
        mode += 2**11 if value[3,1] == "S"
        puts "changing mode of \"#{file}\" to " + sprintf("%04o", mode)
	File.chmod(mode, file)
      else
        p "#{key} not writable"
      end
    end
  end


  # depricated?
  def get_service_type(file)
    reply = @files.GetServiceType(file)
  end


  def list_directory(folder, fields = ["File:Name"])
    ret = {}
    Dir.entries(folder).grep(/^[^.]/).grep(/[^~]$/).each do |f|
      if fields == ["File:Name"]
        ret[folder+"/"+f]= {"File:Name" => {:value => f, :writable => true}}
      else
        ret[folder+"/"+f]= get_metadata(folder + "/" + f, fields)
      end
    end
    ret
  end


  def search_text(query, fields = ["File:Path"])
    reply = @search.Text(-1, "Files", query, 0, 1000, true)
    return reply if fields == ["File:Path"]
    rank = 1 # whats this line for??
    ret = {}
    reply.each do |file|
      data = @metadata.Get("Files", file, fields)
      hash = { "File:Rank" => rank } # whats this line for??
      data.each_index { |i| ret[fields[i]] = data[i] }
      rank += 1
      ret[file] = hash
    end
#    p ret
    ret
#    {"/home/peterpan/Desktop/Keane - (2004) Hopes And Fears"=>{"File.IconPath"=>{:writable=>true, :value=>""}, "File.Name"=>{:writable=>true, :value=>"Keane - (2004) Hopes And Fears"}, "File.Format"=>{:writable=>nil, :value=>"Folder"}, "File.SmallThumbnailPath"=>{:writable=>true, :value=>""}}}
  end


  # depricated?
  def load_path(path, fields = nil)
#    env binding, :path, :fields
    return list_directory(path, fields) if path[0, 1] == "/"
    path =~ /query=\"([^""]+)\"/
    query = $1
#    env binding, "path", "fields", "query"
    return search_text(query, fields) unless query.nil?
    return [] if fields.nil?
    return {}
  end


  def model_for_path(path, show_hidden, columns)
    # TODO: use GetMetadataForFilesInFolder instead of fetching info for each
    #       file. That should speed up things considerably. this is only a hack.
    # This stub fills liststore with the following columns
    # path, type (0=dir 1=file), str (defined by block), pixbuf (defined by
    # block), filename, mimetype, filesize, mtime)
    col_types = {:path => String, :type => Integer, "File:Name" => String,
                 "File:Type" => String, "File:Size" => Integer,
                 "File:Modified" => Time}
    col_types.default= String
    types = columns.collect {|col| (col.class == Class ? col : col_types[col]) }
    model = Gtk::ListStore.new(*types)
    elements = Dir.entries(path) - [".", ".."]
    elements = elements.grep(/^[^.]/) unless show_hidden
    elements.each() do |fname|
      iter = model.append
      fpath = "#{path}/#{fname}"
      i = 0
      iter[i] = fpath unless (i = columns.index :path).nil?
      iter[i] = fname unless (i = columns.index "File:Name").nil?
      stat = File.stat fpath
      if stat.directory?
        iter[i] = 0 unless (i = columns.index :type).nil?
        iter[i] = "directory" unless (i = columns.index "File:Mime").nil?
      else
        iter[i] = 1 unless (i = columns.index :type).nil?
        iter[i] = "file" unless (i = columns.index "File:Mime").nil?
      end
      iter[i] = stat.size unless (i = columns.index "File:Size").nil?
      iter[i] = stat.mtime unless (i = columns.index "File:Modified").nil?
      
      tracker_columns = columns - [:path, :type, "File:Name", "File:Size",
                                   "File:Modified", "File:Mime"]
      tracker_columns = tracker_columns.delete_if { |col| col.class == Class }
      p tracker_columns
      unless tracker_columns.empty?
        tracker_data = get_metadata(fpath, tracker_columns)
        tracker_data.each_pair do |metatype, data|
          data[:value] = "directory" if metatype == "File:Mime" and
                                        data[:value] == "Folder"
          iter[columns.index metatype] = data[:value]
        end
      end      

      yield iter if block_given?
    end
    model
  end
end


$tracker = Tracker.new



