## Meta Sepia
#
# views.rb
# provides the 3 basic views for showing files
# - details-view shows filename and details in columns
# - icon-view shows icons or miniatures
# - list-view shows shows icon and filename side-by-side
#
# author::    Peter Rullmann (mailto:peter@p4n.net)
# copyright:: Copyright (c) 2006-2008 Peter Rullmann
#
# You can Freely distribute this program under the GNU General Public
# License. See the file "COPYING" for the exact licensing terms.

require 'gtk2'
require 'cgi'
#require 'tracker.rb'
require 'location.rb'
require 'digest/md5'
require 'tracker.rb'


#   # #   # #    ##### ### #   # ### #### #   #
## ## #   # #      #    #  #   #  #  #    #   #
# # # #   # #      #    #  #   #  #  ##   #   #
#   # #   # #      #    #   # #   #  #    # # #
#   #  ###  ####   #   ###   #   ### ####  # #
class MultiView < Gtk::HBox
  type_register
  signal_new("sel_changed", GLib::Signal::RUN_FIRST, nil, nil, Array)
  def signal_do_sel_changed(ary); end
  
  def initialize()
    super()
    
    # create views
    @views = {}
    @views[:details] = DetailsView.new
    @views[:icon]    = IconView.new
    @views[:list]    = ListView.new
    
    @views.each do |name, view|
      pack_start view
      view.visible= false
      #view.no_show_all= true

      # forward signals from views to self
      view.signal_connect("path_clicked") do |se, path|
        # TODO: if file selected, then open file...
        $location.path= path
      end      
      view.signal_connect("sel_changed") do |se, files|
        signal_emit("sel_changed", files)
      end
    end

    $glade['zoom_button'].signal_connect("clicked") do |butt|
      $glade['zoom_menu'].popup(nil, nil, 3, 0) do |menu, x, y, push_in|
        butt_alloc = $glade['zoom_button'].allocation
        win_x, win_y = butt.window.position
        [butt_alloc.x + butt_alloc.width - menu.size_request[0] + win_x,
         butt_alloc.y + butt_alloc.height + win_y]
      end
    end

    init_actions
  end


  def init_actions
    # connect to location changes
    $location.signal_connect("changed") do
      # view options should not have been changed
      # model should be up to date
      # sorting should be up to date
      # scroll position should be set as it was
      @views[$location.loc_obj.view_type].refresh
      # show correct view
      set_active_view $location.loc_obj.view_type
    end

    # connect to view changing combobox
    $glade["set_view_combobox"].signal_connect("changed") do |se|
      view_type = {0 => :details, 1 => :list, 2 => :icon}[se.active]
      $location.loc_obj.view_type = view_type
      # model has to be rebuilt
      # thus, sorting has also to be reapplied
      # scroll position can be set to where it was in other view
      @views[$location.loc_obj.view_type].reload
      # change view to chosen view
      set_active_view view_type
    end

    # connect to view changing menu items/actions
    [["details_view_action", :details], ["list_view_action", :list],
     ["icon_view_action", :icon]].each do |action, view_type|
      $glade[action].signal_connect("activate") do |se|
        if $glade[action].active?
          $location.loc_obj.view_type = view_type
          # model has to be rebuilt
          # thus, sorting has also to be reapplied
          # scroll position can be set to where it was in other view
          @views[$location.loc_obj.view_type].reload
          # change view to chosen view
          set_active_view view_type
        end
      end
    end

    # connect to some toggle menu items/actions
    $glade["folders_first_action"].signal_connect("toggled") do |se|
      $location.loc_obj.folders_first= se.active?
      # model is ok
      # only sorting has to be reapplied
      # scroll position needs to be reset
      @active_view.resort
    end
    $glade["reverse_sort_action"].signal_connect("toggled") do |se|
      $location.loc_obj.reverse_sort= se.active?
      # model is ok
      # only sorting has to be reapplied
      # scroll position needs to be reset
      @active_view.resort
    end
    $glade["show_hidden_action"].signal_connect("toggled") do |se|
      $location.loc_obj.show_hidden= se.active?
      # model has to be rebuilt
      # thus, sorting has to be reapplied aswell
      # scroll position can be set as it was
      @active_view.reload
    end

    # connect to menu items/actions for determining sorting order
    # TODO: has to be dynamic not fixed
    [["sort_by_name_action", "File:Name"], ["sort_by_size_action", "File:Size"],
     ["sort_by_type_action", "File:Mime"],
     ["sort_by_date_action", "File:Modified"]].each do |action, sort_by|
      $glade[action].signal_connect("activate") do |se|
        $location.loc_obj.sort_by= sort_by
        # model is ok
        # only sorting has to be reapplied
        # scroll position needs to be reset
        @active_view.resort
      end
    end
    
    # connect to menu items/actions that change the number of lines displayed
    [["0_line_action", 0], ["1_line_action", 1], ["2_lines_action", 2],
     ["4_lines_action", 4], ["6_lines_action", 6]].each do |action, num_lines|
      $glade[action].signal_connect("activate") do |se|
        $location.loc_obj.num_lines= num_lines
        # model has to be rebuilt
        # thus, sorting has also to be reapplied
        # scroll position should be set as it was
        @active_view.reload
      end
    end

    # connect to menu items/actions that change the icon size
    [["small_icons_action", :small], ["medium_icons_action", :medium],
     ["large_icons_action", :large]].each do |action, icon_size|
      $glade[action].signal_connect("activate") do |se|
        $location.loc_obj.icon_size= icon_size
        # model has to be rebuilt
        # thus, sorting has also to be reapplied
        # scroll position should be set as it was
        @active_view.reload
      end
    end
  end

  def set_active_view(view_type)
    # show current view, hide the others
    @views.each do |key, view|
      if key == view_type
        @active_view = view
        view.visible= true
      else
        view.visible= false
      end
    end
    # show only settings which make sense for this view
    ["small_icons_action", "medium_icons_action", "large_icons_action",
     "0_line_action", "1_line_action", "2_lines_action", "4_lines_action",
     "6_lines_action", "group_submenu", "show_columns_submenu"].each do |action|
      $glade[action].sensitive= !@active_view.active_actions.index(action).nil?
    end
    # let view menu reflect the settings from current location object
    curr_obj = $location.loc_obj
    [["icon_view_action", view_type == :icon],
     ["list_view_action", view_type == :list],
     ["details_view_action", view_type == :details],
     ["small_icons_action", curr_obj.icon_size == :small],
     ["medium_icons_action", curr_obj.icon_size == :medium],
     ["large_icons_action", curr_obj.icon_size == :large],
     ["0_line_action", curr_obj.num_lines == 0],
     ["1_line_action", curr_obj.num_lines == 1],
     ["2_lines_action", curr_obj.num_lines == 2],
     ["4_lines_action", curr_obj.num_lines == 4],
     ["6_lines_action", curr_obj.num_lines == 6],
     ["sort_by_name_action", curr_obj.sort_by == "File:Name"],
     ["sort_by_date_action", curr_obj.sort_by == "File:Modified"],
     ["sort_by_size_action", curr_obj.sort_by == "File:Size"],
     ["sort_by_type_action", curr_obj.sort_by == "File:Type"],
     ["show_hidden_action", curr_obj.show_hidden],
     ["folders_first_action", curr_obj.folders_first],
     ["reverse_sort_action", curr_obj.reverse_sort]].each do |menu_entry, bool|
      unless $glade[menu_entry].active? == bool
        $glade[menu_entry].active= bool
      end
    end
    # let combobox reflect current view
    idx = {:details => 0, :list => 1, :icon => 2}[view_type]
    unless $glade["set_view_combobox"].active == idx
      $glade["set_view_combobox"].active= idx
    end
  end
  
end









=begin

  def refresh
    self.count.times { self.take_item(0) }

    fields = ["File:Name", "File:Pixmap"]
    deps = (fields.collect { |a| $typesystem.attributes[a][:deps] }).flatten
    result = $tracker.load_path(@path, deps)

    fmetrics = self.font_metrics
    0.upto(count - 1) { |n| take_item(n) }
    result.keys.sort.each do |file|
      fname = result[file]["File:Name"][:value]
      # compute line wrapping and put lines in fname_lines
      fname_lines = [""]
      lastwidth = 0; linelen = ZOOM_STAGES[@zoom][:linelen]
      fname.each(' ') do |word|
        width = fmetrics.width(word, -1)
        if (width + lastwidth) > linelen
          if lastwidth > linelen
            # word won't fit in line and has to be split up
            prev_width = 0; next_start = 0
            word.size.times do |i|
              width = fmetrics.width(word, i)
              if width - prev_width > linelen
                fname_lines.push(word[next_start, i - next_start])
                prev_width = fmetrics.width(word, i-1)
                next_start = i
              end
            end
            lastwidth = width - prev_width
            fname_lines.push(word[next_start, word.size - next_start])
          else
            lastwidth = width
            fname_lines.push(word)
          end
        else
          lastwidth = lastwidth + width
          fname_lines.push(fname_lines.pop + word)
        end
      end
       
      # TODO: fill with metadata, normalize icon sizes
      fname = fname_lines.join("\n")
#      case ZOOM_STAGES[@zoom][:lines] - fname_lines.size
#      when 1 then fname = fname_lines.join("\n") + "\nfirst"
#      when 2 then fname = fname_lines.join("\n") + "\nfirst\nsecond"
#      when 3 then fname = fname_lines.join("\n") + "\nfirst\nsecond\nthird"
#      end

      item = Qt::ListWidgetItem.new(fname, self)
      item.text_alignment= Qt::AlignLeft | Qt::TextWordWrap
      item.set_data(Qt::UserRole, Qt::Variant.new(file))
      attr = $typesystem.attributes["File:Pixmap"]
      vals = attr[:deps].collect { |a| result[file][a][:value] }
      disp = attr[:display].call(*vals)
      ttip = attr[:tooltip].call(*vals)
      item.icon= Qt::Icon.new(disp)
    end

  end

=end



#    ###  ### ##### #   # ### #### #   # 
#     #  #      #   #   #  #  #    #   # 
#     #   ##    #   #   #  #  ###  #   # 
#     #     #   #    # #   #  #    # # # 
#### ### ###    #     #   ### ####  # #  
class ListView < Gtk::ScrolledWindow
  type_register
  signal_new("path_clicked", GLib::Signal::RUN_FIRST, nil, nil, String)
  def signal_do_path_clicked(str); end
  signal_new("sel_changed", GLib::Signal::RUN_FIRST, nil, nil, Array)
  def signal_do_sel_changed(ary); end

  SIZE_PER_LINE = 16
  
  def active_actions
    ["1_line_action", "2_lines_action", "4_lines_action", "6_lines_action"]
  end
  
  def initialize
    super
    set_policy Gtk::POLICY_NEVER, Gtk::POLICY_AUTOMATIC
    shadow_type= Gtk::SHADOW_IN

    @view = Gtk::IconView.new
    add @view

    @view.selection_mode= Gtk::SELECTION_MULTIPLE
    @view.orientation= Gtk::ORIENTATION_HORIZONTAL
    @view.reorderable= false
    #@view.margin= 0
    @view.spacing= 3
    #@view.column_spacing= 0
    @view.row_spacing= 0
    @view.width_request= 100
    
    renderer = Gtk::CellRendererPixbuf.new
    @view.pack_start(renderer, false)
    @view.set_attributes(renderer, :pixbuf => 3)
    renderer = Gtk::CellRendererText.new
    renderer.ellipsize=Pango::ELLIPSIZE_END
    @view.pack_end(renderer, true)
    @view.set_attributes(renderer, :markup => 2)

    @view.signal_connect("size-allocate") do |se, alloc|
      width = alloc.width
      cols = (width > 300 ? width / 300 : 1)
      col_width = (width - 2 * @view.margin) / cols - @view.spacing * (cols - 1)

      @view.columns = cols
      @view.item_width = col_width
      @size = $location.loc_obj.num_lines * SIZE_PER_LINE + 4
      renderer.width = col_width - 4 * renderer.xpad - @size
    end
    
    @view.signal_connect("item-activated") do |se, path|
      iter = $location.loc_obj.model.get_iter path
      signal_emit("path_clicked", iter[0])
    end
    @view.signal_connect("selection-changed") do |se|
      selection = []
      se.selected_each do |view, path|
        selection << $location.loc_obj.model.get_iter(path)[0]
      end
      signal_emit("sel_changed", selection)
    end
  end

  def reload
    $location.loc_obj.model= nil
    refresh
  end

  def refresh
    loc_obj = $location.loc_obj
    @size = loc_obj.num_lines * SIZE_PER_LINE + 4

    if loc_obj.model.nil?
      icon_gen = IconGenerator.new(@size)
      loc_obj.model_columns = [:path, :type, String, Gdk::Pixbuf, "File:Name",
                               "File:Mime", "File:Size", "File:Modified"]
      loc_obj.model= $tracker.model_for_path(loc_obj.path, loc_obj.show_hidden,
                                             loc_obj.model_columns) do |iter|
        case loc_obj.num_lines
          when 0 then iter[2]= ""
          when 1 then iter[2]= CGI.escapeHTML(iter[4])
          when 2 then iter[2]= "#{CGI.escapeHTML(iter[4])}" +
            "\n<span foreground='grey'>#{DateMetaType.display(iter[7])}</span>"
          when 4 then iter[2]= "#{CGI.escapeHTML(iter[4])}" +
            "\n<span foreground='grey'>#{DateMetaType.display(iter[7])}" +
            "\n#{StringMetaType.display(iter[5])}" +
            "\n#{FilesizeMetaType.display(iter[6])}</span>"
          when 6 then iter[2]= "#{CGI.escapeHTML(iter[4])}" +
            "\n<span foreground='grey'>#{DateMetaType.display(iter[7])}" +
            "\n#{StringMetaType.display(iter[5])}" +
            "\n#{FilesizeMetaType.display(iter[6])}\n5th row\n6th row</span>"
          end
        iter[3]= icon_gen.icon(iter[0], iter[5])
      end
    end
    resort_helper
    @view.model= loc_obj.model
    rescroll
  end
  
  def resort
    resort_helper
    $location.loc_obj.scroll_point= 0
    rescroll
  end
  
  def resort_helper
    loc_obj = $location.loc_obj
    sort_col = loc_obj.model_columns.index loc_obj.sort_by
    order = (loc_obj.reverse_sort ? Gtk::SORT_DESCENDING : Gtk::SORT_ASCENDING)
    loc_obj.model.set_sort_column_id sort_col, order
    loc_obj.model.set_sort_column_id 1 if loc_obj.folders_first
  end
  
  def rescroll
    scroll_point = $location.loc_obj.scroll_point
    adj = vadjustment
    adj.value= scroll_point * (adj.upper - adj.lower) / 1000 + adj.lower
  end
end




###  #### #####  ##  ### #     ### #   # ### #### #   # 
#  # #      #   #  #  #  #    #    #   #  #  #    #   # 
#  # ###    #   ####  #  #     ##  #   #  #  ###  #   # 
#  # #      #   #  #  #  #       #  # #   #  #    # # # 
###  ####   #   #  # ### #### ###    #   ### ####  # #  
class DetailsView < Gtk::ScrolledWindow
  type_register
  signal_new("path_clicked", GLib::Signal::RUN_FIRST, nil, nil, String)
  def signal_do_path_clicked(str); end
  signal_new("sel_changed", GLib::Signal::RUN_FIRST, nil, nil, Array)
  def signal_do_sel_changed(ary); end

  ICON_SIZE= 18
  
  def active_actions
    []
  end
  
  def initialize
    super
    set_policy Gtk::POLICY_NEVER, Gtk::POLICY_AUTOMATIC
    shadow_type= Gtk::SHADOW_IN

    @view = Gtk::TreeView.new
    add @view

    @view.selection.mode= Gtk::SELECTION_MULTIPLE
    @view.reorderable= false
    @view.headers_visible= true
    @view.headers_clickable= true
    @view.rules_hint= true
    #@view.margin= 0
    
    @view.signal_connect("row-activated") do |se, path, col|
      iter = $location.loc_obj.model.get_iter path
      signal_emit("path_clicked", iter[0])
    end
    @view.selection.signal_connect("changed") do |se|
      selection = []
      se.selected_each do |view, path|
        selection << $location.loc_obj.model.get_iter(path)[0]
      end
      signal_emit("sel_changed", selection)
    end
    
    recolumn
  end

  def recolumn
    @view.columns.each { |col| @view.remove_column(col) }
    col = Gtk::TreeViewColumn.new("Filename")
    col.pack_start(crp = Gtk::CellRendererPixbuf.new, false)
    col.set_attributes(crp, :pixbuf => 3)
    col.pack_start(crt = Gtk::CellRendererText.new, false)
    col.set_attributes(crt, :text => 4)
    @view.append_column(col)
    col = Gtk::TreeViewColumn.new("Size")
    col.pack_start(crt = Gtk::CellRendererText.new, false)
    col.set_attributes(crt, :text => 2)
    @view.append_column(col)
  end

  def reload
    $location.loc_obj.model= nil
    refresh
  end

  def refresh
    loc_obj = $location.loc_obj
    @size = ICON_SIZE

    if loc_obj.model.nil?
      icon_gen = IconGenerator.new(@size)
      loc_obj.model_columns = [:path, :type, String, Gdk::Pixbuf, "File:Name",
                               "File:Mime", "File:Size", "File:Modified"]
      loc_obj.model= $tracker.model_for_path(loc_obj.path, loc_obj.show_hidden,
                                             loc_obj.model_columns) do |iter|
        iter[2]= FilesizeMetaType.display(iter[6])
        iter[3]= icon_gen.icon(iter[0], iter[5])
      end
    end
    resort_helper
    @view.model= loc_obj.model
    rescroll
  end
  
  def resort
    resort_helper
    $location.loc_obj.scroll_point= 0
    rescroll
  end
  
  def resort_helper
    loc_obj = $location.loc_obj
    sort_col = loc_obj.model_columns.index loc_obj.sort_by
    order = (loc_obj.reverse_sort ? Gtk::SORT_DESCENDING : Gtk::SORT_ASCENDING)
    loc_obj.model.set_sort_column_id sort_col, order
    loc_obj.model.set_sort_column_id 1 if loc_obj.folders_first
  end
  
  def rescroll
    scroll_point = $location.loc_obj.scroll_point
    adj = vadjustment
    adj.value= scroll_point * (adj.upper - adj.lower) / 1000 + adj.lower
  end
end


###  ##   ###  #   # #   # ### #### #   # 
 #  #  # #   # ##  # #   #  #  #    #   # 
 #  #    #   # # # # #   #  #  ###  #   # 
 #  #  # #   # #  ##  # #   #  #    # # # 
###  ##   ###  #   #   #   ### ####  # #  
class IconView < Gtk::ScrolledWindow
  type_register
  signal_new("path_clicked", GLib::Signal::RUN_FIRST, nil, nil, String)
  def signal_do_path_clicked(str); end
  signal_new("sel_changed", GLib::Signal::RUN_FIRST, nil, nil, Array)
  def signal_do_sel_changed(ary); end

  ICON_SIZES = {:small => 75, :medium => 130, :large => 250}
  
  def active_actions
    ["small_icons_action", "medium_icons_action", "large_icons_action",
     "0_line_action", "1_line_action", "2_lines_action", "4_lines_action",
     "6_lines_action"]
  end
  
  def initialize
    super
    set_policy Gtk::POLICY_NEVER, Gtk::POLICY_AUTOMATIC
    shadow_type= Gtk::SHADOW_IN

    @view = Gtk::IconView.new
    add @view

    @view.selection_mode= Gtk::SELECTION_MULTIPLE
    @view.orientation= Gtk::ORIENTATION_VERTICAL
    @view.reorderable= false
    #@view.margin= 0
    @view.spacing= 3
    #@view.column_spacing= 0
    @view.row_spacing= 0
    
    renderer = Gtk::CellRendererPixbuf.new
    @view.pack_start(renderer, false)
    @view.set_attributes(renderer, :pixbuf => 3)
    @text_renderer = Gtk::CellRendererText.new
    @text_renderer.wrap_mode= Pango::WRAP_CHAR
    @text_renderer.alignment= Pango::ALIGN_CENTER
    @view.pack_end(@text_renderer, true)
    @view.set_attributes(@text_renderer, :markup => 2)
    
    @view.signal_connect("item-activated") do |se, path|
      iter = $location.loc_obj.model.get_iter path
      signal_emit("path_clicked", iter[0])
    end
    @view.signal_connect("selection-changed") do |se|
      selection = []
      se.selected_each do |view, path|
        selection << $location.loc_obj.model.get_iter(path)[0]
      end
      signal_emit("sel_changed", selection)
    end
  end

  def reload
    $location.loc_obj.model= nil
    refresh
  end

  def refresh
    loc_obj = $location.loc_obj
    @size = ICON_SIZES[ loc_obj.icon_size ]
    @view.item_width= [@size + 5, (loc_obj.num_lines == 0 ? 0 : 150)].max
    @text_renderer.wrap_width= @view.item_width

    if loc_obj.model.nil?
      icon_gen = IconGenerator.new(@size)
      loc_obj.model_columns = [:path, :type, String, Gdk::Pixbuf, "File:Name",
                               "File:Mime", "File:Size", "File:Modified"]
      loc_obj.model= $tracker.model_for_path(loc_obj.path, loc_obj.show_hidden,
                                             loc_obj.model_columns) do |iter|
        case loc_obj.num_lines
          when 0 then iter[2]= ""
          when 1 then iter[2]= CGI.escapeHTML(iter[4])
          when 2 then iter[2]= "#{iter[4]}\n<span foreground='grey'>" +
            DateMetaType.display(iter[7]) + "</span>"
          when 3..6 then iter[2]= "#{iter[4]}\n<span foreground='grey'>" +
            DateMetaType.display(iter[7]) + "\n" + 
            FilesizeMetaType.display(iter[6]) + "</span>"
          end
        iter[3]= icon_gen.icon(iter[0], iter[5])
      end
    end
    resort_helper
    @view.model= loc_obj.model
    rescroll
  end
  
  def resort
    resort_helper
    $location.loc_obj.scroll_point= 0
    rescroll
  end
  
  def resort_helper
    loc_obj = $location.loc_obj
    sort_col = loc_obj.model_columns.index loc_obj.sort_by
    order = (loc_obj.reverse_sort ? Gtk::SORT_DESCENDING : Gtk::SORT_ASCENDING)
    loc_obj.model.set_sort_column_id sort_col, order
    loc_obj.model.set_sort_column_id 1 if loc_obj.folders_first
  end
  
  def rescroll
    scroll_point = $location.loc_obj.scroll_point
    adj = vadjustment
    adj.value= scroll_point * (adj.upper - adj.lower) / 1000 + adj.lower
  end
end




###  ##   ###  #   #  ##  #### #   # #### ###   ##  #####  ###  ###  
 #  #  # #   # ##  # #    #    ##  # #    #  # #  #   #   #   # #  # 
 #  #    #   # # # # # ## ###  # # # ###  ###  ####   #   #   # ###  
 #  #  # #   # #  ## #  # #    #  ## #    #  # #  #   #   #   # #  # 
###  ##   ###  #   #  ### #### #   # #### #  # #  #   #    ###  #  # 
class IconGenerator
  attr_reader :file_icon, :folder_icon

  def initialize(size)
    @size = size
    @base_dir = ENV["HOME"] + "/.thumbnails"

    # load or generate default icons
    @icon_theme = Gtk::IconTheme.default
    @file_icon = @icon_theme.load_icon("text-x-preview", size,
                                       Gtk::IconTheme::LOOKUP_FORCE_SVG)
    @folder_icon = @icon_theme.load_icon("folder", size,
                                         Gtk::IconTheme::LOOKUP_FORCE_SVG)
    @blank_icon = @file_icon.scale(size, size)
    @blank_icon.fill!(0xffffffff)
    @black_icon = @blank_icon.dup.fill!(0x000000ff)
    @grey_icon = @blank_icon.dup.fill!(0xbbbbbbff)
  end
  
  def icon(path, mime_type)
    md5 = Digest::MD5.hexdigest "file://#{path}"
    thumb = "/home/peterpan/.thumbnails/normal/#{md5}.png"
    if File.exists? thumb
      th = Gdk::Pixbuf.new(thumb, @size, @size)
      blank = @blank_icon.dup
      xoff = (@size - th.width) / 2
      yoff = (@size - th.height) / 2
      @black_icon.copy_area(0,0, th.width, th.height, blank, xoff, yoff)
      @grey_icon.copy_area(1,1, th.width-2, th.height-2, blank, xoff+1, yoff+1)
      th.copy_area(2,2, th.width-4, th.height-4, blank, xoff+2, yoff+2)
      return blank
    else
      # TODO: test for .thumbnails/fail/....png and generate thumb if not
      # TODO: support more mime types
      if mime_type =~ /^directory/
        return @folder_icon
      else
        return @file_icon
      end
    end
  end

end





